@@ -22,6 +22,7 @@
 
 #include <iconv.h>
 #include <errno.h>
+#include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
 
@@ -187,30 +188,264 @@ g_iconv_close (GIConv converter)
   return iconv_close (cd);
 }
 
+
+#define ICONV_CACHE_SIZE   (16)
+
+struct _iconv_cache_bucket {
+  gchar *key;
+  guint32 refcount;
+  gboolean used;
+  iconv_t cd;
+};
+
+static GList *iconv_cache_list;
+static GHashTable *iconv_cache;
+static GHashTable *iconv_open_hash;
+static guint iconv_cache_size = 0;
+G_LOCK_DEFINE_STATIC (iconv_cache_lock);
+
+/* caller *must* hold the iconv_cache_lock */
+static void
+iconv_cache_init (void)
+{
+  static gboolean initialized = FALSE;
+  
+  if (initialized)
+    return;
+  
+  iconv_cache_list = NULL;
+  iconv_cache = g_hash_table_new (g_str_hash, g_str_equal);
+  iconv_open_hash = g_hash_table_new (g_direct_hash, g_direct_equal);
+  
+  initialized = TRUE;
+}
+
+
+/**
+ * iconv_cache_bucket_new:
+ * @key: cache key
+ * @cd: iconv descriptor
+ *
+ * Creates a new cache bucket, inserts it into the cache and
+ * increments the cache size.
+ *
+ * Returns a pointer to the newly allocated cache bucket.
+ **/
+struct _iconv_cache_bucket *
+iconv_cache_bucket_new (const gchar *key, iconv_t cd)
+{
+  struct _iconv_cache_bucket *bucket;
+  
+  bucket = g_new (struct _iconv_cache_bucket, 1);
+  bucket->key = g_strdup (key);
+  bucket->refcount = 1;
+  bucket->used = TRUE;
+  bucket->cd = cd;
+  
+  g_hash_table_insert (iconv_cache, bucket->key, bucket);
+  
+  /* FIXME: if we sorted the list so items with few refcounts were
+     first, then we could expire them faster in iconv_cache_expire_unused () */
+  iconv_cache_list = g_list_prepend (iconv_cache_list, bucket);
+  
+  iconv_cache_size++;
+  
+  return bucket;
+}
+
+
+/**
+ * iconv_cache_bucket_expire:
+ * @node: cache bucket's node
+ * @bucket: cache bucket
+ *
+ * Expires a single cache bucket @bucket. This should only ever be
+ * called on a bucket that currently has no used iconv descriptors
+ * open.
+ *
+ * @node is not a required argument. If @node is not supplied, we
+ * search for it ourselves.
+ **/
+static void
+iconv_cache_bucket_expire (GList *node, struct _iconv_cache_bucket *bucket)
+{
+  g_hash_table_remove (iconv_cache, bucket->key);
+  
+  if (node == NULL)
+    node = g_list_find (iconv_cache_list, bucket);
+  
+  g_assert (node != NULL);
+  
+  if (node->prev)
+    {
+      node->prev->next = node->next;
+      if (node->next)
+        node->next->prev = node->prev;
+    }
+  else
+    {
+      iconv_cache_list = node->next;
+      if (node->next)
+        node->next->prev = NULL;
+    }
+  
+  g_list_free_1 (node);
+  
+  g_free (bucket->key);
+  g_iconv_close (bucket->cd);
+  g_free (bucket);
+  
+  iconv_cache_size--;
+}
+
+
+/**
+ * iconv_cache_expire_unused:
+ *
+ * Expires as many unused cache buckets as it needs to in order to get
+ * the total number of buckets < ICONV_CACHE_SIZE.
+ **/
+static void
+iconv_cache_expire_unused (void)
+{
+  struct _iconv_cache_bucket *bucket;
+  GList *node, *next;
+  
+  node = iconv_cache_list;
+  while (node && iconv_cache_size >= ICONV_CACHE_SIZE)
+    {
+      next = node->next;
+      
+      bucket = node->data;
+      if (bucket->refcount == 0)
+        iconv_cache_bucket_expire (node, bucket);
+      
+      node = next;
+    }
+}
+
 static GIConv
 open_converter (const gchar *to_codeset,
-                const gchar *from_codeset,
+		const gchar *from_codeset,
 		GError     **error)
 {
-  GIConv cd = g_iconv_open (to_codeset, from_codeset);
-
-  if (cd == (iconv_t) -1)
+  struct _iconv_cache_bucket *bucket;
+  gchar *key;
+  GIConv cd;
+  
+  /* create our key */
+  key = g_alloca (strlen (from_codeset) + strlen (to_codeset) + 2);
+  sprintf (key, "%s:%s", from_codeset, to_codeset);
+  
+  G_LOCK (iconv_cache_lock);
+  
+  /* make sure the cache has been initialized */
+  iconv_cache_init ();
+  
+  bucket = g_hash_table_lookup (iconv_cache, key);
+  if (bucket)
     {
-      /* Something went wrong.  */
-      if (errno == EINVAL)
-        g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_NO_CONVERSION,
-                     _("Conversion from character set '%s' to '%s' is not supported"),
-                     from_codeset, to_codeset);
+      if (bucket->used)
+        {
+          cd = g_iconv_open (to_codeset, from_codeset);
+          if (cd == (iconv_t) -1)
+            goto error;
+        }
       else
-        g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
-                     _("Could not open converter from '%s' to '%s': %s"),
-                     from_codeset, to_codeset, strerror (errno));
+        {
+          cd = bucket->cd;
+          bucket->used = TRUE;
+          
+          /* reset the descriptor */
+          g_iconv (cd, NULL, NULL, NULL, NULL);
+        }
+      
+      bucket->refcount++;
     }
-
+  else
+    {
+      cd = g_iconv_open (to_codeset, from_codeset);
+      if (cd == (iconv_t) -1)
+        goto error;
+      
+      iconv_cache_expire_unused ();
+      
+      bucket = iconv_cache_bucket_new (key, cd);
+    }
+  
+  g_hash_table_insert (iconv_open_hash, cd, bucket->key);
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  return cd;
+  
+ error:
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  /* Something went wrong.  */
+  if (errno == EINVAL)
+    g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_NO_CONVERSION,
+		 _("Conversion from character set '%s' to '%s' is not supported"),
+		 from_codeset, to_codeset);
+  else
+    g_set_error (error, G_CONVERT_ERROR, G_CONVERT_ERROR_FAILED,
+		 _("Could not open converter from '%s' to '%s': %s"),
+		 from_codeset, to_codeset, strerror (errno));
+  
   return cd;
-
 }
 
+static int
+close_converter (GIConv converter)
+{
+  struct _iconv_cache_bucket *bucket;
+  const gchar *key;
+  iconv_t cd;
+  
+  cd = (iconv_t) converter;
+  
+  if (cd == (iconv_t) -1)
+    return 0;
+  
+  G_LOCK (iconv_cache_lock);
+  
+  key = g_hash_table_lookup (iconv_open_hash, cd);
+  if (key)
+    {
+      g_hash_table_remove (iconv_open_hash, cd);
+      
+      bucket = g_hash_table_lookup (iconv_cache, key);
+      g_assert (bucket);
+      
+      bucket->refcount--;
+      
+      if (cd == bucket->cd)
+        bucket->used = FALSE;
+      else
+        g_iconv_close (cd);
+      
+      if (!bucket->refcount && iconv_cache_size > ICONV_CACHE_SIZE)
+        {
+          /* expire this cache bucket */
+          iconv_cache_bucket_expire (NULL, bucket);
+        }
+    }
+  else
+    {
+      G_UNLOCK (iconv_cache_lock);
+      
+      g_warning ("This iconv context wasn't opened using open_converter");
+      
+      return g_iconv_close (converter);
+    }
+  
+  G_UNLOCK (iconv_cache_lock);
+  
+  return 0;
+}
+
+
 /**
  * g_convert:
  * @str:           the string to convert
@@ -251,7 +486,7 @@ g_convert (const gchar *str,
   g_return_val_if_fail (str != NULL, NULL);
   g_return_val_if_fail (to_codeset != NULL, NULL);
   g_return_val_if_fail (from_codeset != NULL, NULL);
-     
+  
   cd = open_converter (to_codeset, from_codeset, error);
 
   if (cd == (GIConv) -1)
@@ -269,7 +504,7 @@ g_convert (const gchar *str,
 			      bytes_read, bytes_written,
 			      error);
   
-  g_iconv_close (cd);
+  close_converter (cd);
 
   return res;
 }
@@ -497,7 +732,12 @@ g_convert_with_fallback (const gchar *str,
   utf8 = g_convert (str, len, "UTF-8", from_codeset, 
 		    bytes_read, &inbytes_remaining, error);
   if (!utf8)
-    return NULL;
+    {
+      close_converter (cd);
+      if (bytes_written)
+        *bytes_written = 0;
+      return NULL;
+    }
 
   /* Now the heart of the code. We loop through the UTF-8 string, and
    * whenever we hit an offending character, we form fallback, convert
@@ -594,10 +834,10 @@ g_convert_with_fallback (const gchar *str,
    */
   *outp = '\0';
   
-  g_iconv_close (cd);
+  close_converter (cd);
 
   if (bytes_written)
-    *bytes_written = outp - str;	/* Doesn't include '\0' */
+    *bytes_written = outp - dest;	/* Doesn't include '\0' */
 
   g_free (utf8);
 
@@ -618,6 +858,8 @@ g_convert_with_fallback (const gchar *str,
  * 
  */
 
+#ifndef G_PLATFORM_WIN32
+
 static gchar *
 strdup_len (const gchar *string,
 	    gssize       len,
@@ -658,6 +900,8 @@ strdup_len (const gchar *string,
   return g_strndup (string, real_len);
 }
 
+#endif
+
 /**
  * g_locale_to_utf8:
  * @opsysstring:   a string in the encoding of the current locale
@@ -1064,13 +1308,19 @@ typedef enum {
 } UnsafeCharacterSet;
 
 static const guchar acceptable[96] = {
- /* X0   X1   X2   X3   X4   X5   X6   X7   X8   X9   XA   XB   XC   XD   XE   XF */
-  0x00,0x3F,0x20,0x20,0x20,0x00,0x2C,0x3F,0x3F,0x3F,0x3F,0x22,0x20,0x3F,0x3F,0x1C, /* 2X  !"#$%&'()*+,-./   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x38,0x20,0x20,0x2C,0x20,0x2C, /* 3X 0123456789:;<=>?   */
-  0x30,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F, /* 4X @ABCDEFGHIJKLMNO   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x20,0x3F, /* 5X PQRSTUVWXYZ[\]^_   */
-  0x20,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F, /* 6X `abcdefghijklmno   */
-  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x3F,0x20  /* 7X pqrstuvwxyz{|}~DEL */
+  /* A table of the ASCII chars from space (32) to DEL (127) */
+  /*      !    "    #    $    %    &    '    (    )    *    +    ,    -    .    / */ 
+  0x00,0x3F,0x20,0x20,0x20,0x00,0x2C,0x3F,0x3F,0x3F,0x3F,0x22,0x20,0x3F,0x3F,0x1C,
+  /* 0    1    2    3    4    5    6    7    8    9    :    ;    <    =    >    ? */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x38,0x20,0x20,0x2C,0x20,0x2C,
+  /* @    A    B    C    D    E    F    G    H    I    J    K    L    M    N    O */
+  0x30,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,
+  /* P    Q    R    S    T    U    V    W    X    Y    Z    [    \    ]    ^    _ */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x20,0x3F,
+  /* `    a    b    c    d    e    f    g    h    i    j    k    l    m    n    o */
+  0x20,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,
+  /* p    q    r    s    t    u    v    w    x    y    z    {    |    }    ~  DEL */
+  0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x3F,0x20,0x20,0x20,0x3F,0x20
 };
 
 static const gchar hex[16] = "0123456789ABCDEF";
@@ -1101,7 +1351,7 @@ g_escape_uri_string (const gchar *string,
   use_mask = mask;
   for (p = string; *p != '\0'; p++)
     {
-      c = *p;
+      c = (guchar) *p;
       if (!ACCEPTABLE (c)) 
 	unacceptable++;
     }
@@ -1111,7 +1361,7 @@ g_escape_uri_string (const gchar *string,
   use_mask = mask;
   for (q = result, p = string; *p != '\0'; p++)
     {
-      c = (unsigned char)*p;
+      c = (guchar) *p;
       
       if (!ACCEPTABLE (c))
 	{
@@ -1137,6 +1387,23 @@ g_escape_file_uri (const gchar *hostname,
   char *escaped_path;
   char *res;
 
+#ifdef G_OS_WIN32
+  char *p, *backslash;
+
+  /* Turn backslashes into forward slashes. That's what Netscape
+   * does, and they are actually more or less equivalent in Windows.
+   */
+  
+  pathname = g_strdup (pathname);
+  p = (char *) pathname;
+  
+  while ((backslash = strchr (p, '\\')) != NULL)
+    {
+      *backslash = '/';
+      p = backslash + 1;
+    }
+#endif
+
   if (hostname && *hostname != '\0')
     {
       escaped_hostname = g_escape_uri_string (hostname, UNSAFE_HOST);
@@ -1150,6 +1417,10 @@ g_escape_file_uri (const gchar *hostname,
 		     escaped_path,
 		     NULL);
 
+#ifdef G_OS_WIN32
+  g_free ((char *) pathname);
+#endif
+
   g_free (escaped_hostname);
   g_free (escaped_path);
   
@@ -1251,6 +1522,9 @@ g_filename_from_uri (const char *uri,
   char *result;
   char *filename;
   int offs;
+#ifdef G_OS_WIN32
+  char *p, *slash;
+#endif
 
   if (hostname)
     *hostname = NULL;
@@ -1315,11 +1589,39 @@ g_filename_from_uri (const char *uri,
       return NULL;
     }
 
-  /* DOS uri's are like "file://host/c:\foo", so we need to check if we need to
-   * drop the initial slash */
   offs = 0;
-  if (g_path_is_absolute (filename+1))
-    offs = 1;
+#ifdef G_OS_WIN32
+  /* Drop localhost */
+  if (hostname && *hostname != NULL &&
+      g_ascii_strcasecmp (*hostname, "localhost") == 0)
+    {
+      g_free (*hostname);
+      *hostname = NULL;
+    }
+
+  /* Turn slashes into backslashes, because that's the canonical spelling */
+  p = filename;
+  while ((slash = strchr (p, '/')) != NULL)
+    {
+      *slash = '\\';
+      p = slash + 1;
+    }
+
+  /* Windows URIs with a drive letter can be like "file://host/c:/foo"
+   * or "file://host/c|/foo" (some Netscape versions). In those cases, start
+   * the filename from the drive letter.
+   */
+  if (g_ascii_isalpha (filename[1]))
+    {
+      if (filename[2] == ':')
+	offs = 1;
+      else if (filename[2] == '|')
+	{
+	  filename[2] = ':';
+	  offs = 1;
+	}
+    }
+#endif
   
   result = g_filename_from_utf8 (filename + offs, -1, NULL, NULL, error);
   g_free (filename);
@@ -1371,6 +1673,12 @@ g_filename_to_uri   (const char *filename,
       return NULL;
     }
   
+#ifdef G_OS_WIN32
+  /* Don't use localhost unnecessarily */
+  if (hostname && g_ascii_strcasecmp (hostname, "localhost") == 0)
+    hostname = NULL;
+#endif
+
   escaped_uri = g_escape_file_uri (hostname,
 				   utf8_filename);
   g_free (utf8_filename);
